from typing import TypeVar

from pydantic import BaseModel

from ..enums import CacheTypes


class BaseDBO(BaseModel):
    __cache_type__: CacheTypes = CacheTypes.NONE
    __cache_expiration__: int = 0

    @classmethod
    def get_fields(cls) -> list[str]:
        return [ field_name for field_name, _ in cls.model_fields.items() ]

    def __hash__(self) -> int:
        return hash(self.model_dump_json())


ModelDBOType = TypeVar('ModelDBOType', bound=BaseDBO)



from datetime import datetime
from typing import Any

from pydantic import BaseModel

from ..enums import ConditionTypes


class Condition(BaseModel):
    type: ConditionTypes
    value: Any


def not_eq(val: Any) -> Condition:
    return Condition(
        type=ConditionTypes.NOT_EQ,
        value=val,
    )


def gt_eq(val: Any) -> Condition:
    return Condition(
        type=ConditionTypes.GT_EQ,
        value=val,
    )


def le_eq(val: Any) -> Condition:
    return Condition(
        type=ConditionTypes.LE_EQ,
        value=val,
    )


def gt(val: Any) -> Condition:
    return Condition(
        type=ConditionTypes.GT,
        value=val,
    )


def le(val: Any) -> Condition:
    return Condition(
        type=ConditionTypes.LE,
        value=val,
    )


def one_of(val: Any) -> Condition:
    return Condition(
        type=ConditionTypes.ONE_OF,
        value=val,
    )



from typing import Any, Type, Dict

from ...dbo.base import ModelDBOType


class BaseDBODriver:
    def __init__(self,
        dsn: str | None = None,
    ):
        self.dsn = dsn

    async def connect(self):
        raise NotImplementedError()
    
    async def disconnect(self):
        raise NotImplementedError()

    async def is_connected(self) -> bool:
        raise NotImplementedError()
    
    async def init_model(self, dbo_cls: ModelDBOType):
        raise NotImplementedError()
    
    async def create_model(self, dbo_cls: ModelDBOType):
        raise NotImplementedError()

    async def save(self, dbo: ModelDBOType) -> ModelDBOType:
        raise NotImplementedError()
    
    async def get(self, dbo_cls: Type[ModelDBOType], attr: str, value: Any) -> ModelDBOType | None:
        raise NotImplementedError()

    async def get_list(
        self,
        dbo_cls: Type[ModelDBOType],
        attrs: Dict[str, Any] | None = None,
        order_by: str = "id",
        order_direction: str = "desc",
        limit: int = 100,
        offset: int = 0,
    ) -> list[ModelDBOType]:
        raise NotImplementedError()
    
    async def delete(self, dbo: ModelDBOType) -> None:
        raise NotImplementedError()
    
    async def get_count(self, dbo_cls: Type[ModelDBOType], attrs: Dict[str, Any] | None = None) -> int:
        raise NotImplementedError()



import uuid

from datetime import datetime
from typing import TypeVar, Annotated, Any

from pydantic import (
    BaseModel,
    UUID4,
    AfterValidator,
    Field,
)


class DTO(BaseModel):
    @classmethod
    def get_fields(cls) -> list[str]:
        return [ field_name for field_name, _ in cls.model_fields.items() ]

class BaseDTO(DTO):
    id: None | UUID4 | Annotated[str, AfterValidator(lambda x: uuid.UUID(x, version=4))] = Field(
        default=None,
        examples=["00000000-0000-0000-0000-000000000000"],
    )
    created_dt: datetime | None = None
    updated_dt: datetime | None = None


class BaseEvent(BaseModel):
    event_type: str


class BaseError(Exception):
    http_status: int
    error_code: str
    description: str


class BaseMetric(BaseModel):
    ts: int = 0
    metric_domain: str
    metric: str
    instance: str | None = None
    value: Any


ModelDTOType = TypeVar('ModelDTOType', bound=BaseDTO)
ModelEventType = TypeVar('ModelEventType', bound=BaseEvent)



from functools import reduce

from datetime import datetime
from typing import Any, Dict, Type
from urllib.parse import urlparse
from uuid import UUID

from rethinkdb import RethinkDB
from rethinkdb.net import Cursor
from ulid import ULID

from .base import BaseDBODriver
from ...dbo import RethinkDBModelDBOType
from ...enums import ConditionTypes
from ...utils.dt import now
from ...utils.conditions import Condition


class RethinkDBDBODriver(BaseDBODriver):
    def __init__(self, dsn = None):
        self.parsed_dsn = urlparse(dsn)

        self._rdb = RethinkDB()

    def connect(self):
        return self._rdb.connect(
            host=self.parsed_dsn.hostname,
            port=int(self.parsed_dsn.port),
            password=self.parsed_dsn.password,
            db=self.parsed_dsn.path.replace("/", ""),
        )

    async def create_model(self, dbo_cls: RethinkDBModelDBOType):
        with self.connect() as conn:
            tablename = dbo_cls.__tablename__

            table_list = self._rdb.table_list().run(conn)
            if tablename not in table_list:
                self._rdb.table_create(tablename).run(conn)

    async def init_model(self, dbo_cls: RethinkDBModelDBOType):
        with self.connect() as conn:
            tablename = dbo_cls.__tablename__
            indexes = dbo_cls.__indexes__

            table_list = self._rdb.table_list().run(conn)
            if tablename in table_list:
                indexes_list = self._rdb.table(tablename).index_list().run(conn)
                for index in indexes:
                    if index not in indexes_list:
                        self._rdb.table(tablename).index_create(index).run(conn)
                self._rdb.table(tablename).index_wait().run(conn)

    async def save(self, dbo: RethinkDBModelDBOType, is_new: bool = False):
        with self.connect() as conn:
            if not dbo.id:
                is_new = True
                dbo.id = ULID().to_uuid4()

            if not dbo.created_dt:
                dbo.created_dt = now()
            
            dbo.updated_dt = now()

            if is_new:
                self._rdb.table(dbo.__tablename__).insert(dbo.model_dump(mode="json")).run(conn)
            else:
                self._rdb.table(dbo.__tablename__).get(str(dbo.id)).replace(dbo.model_dump(mode="json")).run(conn)

            return dbo

    def adapt_value(self, value: Any) -> Any:
        if isinstance(value, UUID):
            return str(value)
        else:
            return value
        
    def create_filter(self, attrs: dict):
        def query_filter(row: Any):
            expr = self._rdb.expr(True)
            if attrs and len(attrs.keys()) > 0:
                for key in attrs.keys():
                    if isinstance(attrs[key], Condition):
                        key_value: Condition = attrs[key]
                        if isinstance(key_value.value, datetime):
                            filter_field = self._rdb.iso8601(row[key])
                        else:
                            filter_field = row[key]
                        if key_value.type == ConditionTypes.NOT_EQ:
                            expr = expr & (filter_field != self.adapt_value(key_value.value))
                        elif key_value.type == ConditionTypes.GT:
                            expr = expr & (filter_field > self.adapt_value(key_value.value))
                        elif key_value.type == ConditionTypes.GT_EQ:
                            expr = expr & (filter_field >= self.adapt_value(key_value.value))
                        elif key_value.type == ConditionTypes.LE:
                            expr = expr & (filter_field < self.adapt_value(key_value.value))
                        elif key_value.type == ConditionTypes.LE_EQ:
                            expr = expr & (filter_field <= self.adapt_value(key_value.value))
                        elif key_value.type == ConditionTypes.ONE_OF:
                            expr = expr & (self._rdb.expr(key_value.value).contains(row[key]))
                    elif isinstance(attrs[key], str):
                        expr = expr & (row[key] == attrs[key])
                    else:
                        if isinstance(attrs[key], datetime):
                            filter_field = self._rdb.iso8601(row[key])
                        else:
                            filter_field = row[key]
                        expr = expr & (filter_field == attrs[key])
            return expr
        return query_filter


    async def get(self, dbo_cls: RethinkDBModelDBOType, attr: str, value: Any) -> RethinkDBModelDBOType | None:
        with self.connect() as conn:
            tablename = dbo_cls.__tablename__
            indexes = dbo_cls.__indexes__

            query = self._rdb.table(tablename)
            attr_value = self.adapt_value(value)

            if attr == "id":
                query = query.get(attr_value)
            elif attr in indexes:
                query = query.get_all(attr_value, index = attr)
            else:
                query = query.filter(self._rdb.row[attr] == attr_value)

            result = query.run(conn)

            if isinstance(result, Cursor):
                result_doc = None
                for doc in result:
                    result_doc = doc
                    break
                result = result_doc

            if result is None:
                return result

            return dbo_cls.model_validate(result)
    
    async def get_list(
        self,
        dbo_cls: RethinkDBModelDBOType,
        attrs: Dict[str, Any] | None = None,
        order_by: str = "id",
        order_direction: str = "desc",
        limit: int = 100,
        offset: int = 0,
    ) -> list[RethinkDBModelDBOType]:
        with self.connect() as conn:
            tablename = dbo_cls.__tablename__
            result_items: list[RethinkDBModelDBOType] = []

            query = self._rdb.table(tablename)
            adapt_attrs = {}
            if attrs:
                for key in attrs.keys():
                    adapt_attrs[key] = self.adapt_value(attrs[key])

            query = query.filter(self.create_filter(adapt_attrs))
            if order_direction == "desc":
                query = query.order_by(self._rdb.desc(order_by))
            else:
                query = query.order_by(self._rdb.asc(order_by))
            query = query.skip(offset).limit(limit)

            cursor = query.run(conn)
            for item in list(cursor):
                result_items.append(dbo_cls.model_validate(item))

            return result_items

    async def delete(self, dbo: RethinkDBModelDBOType):
        with self.connect() as conn:
            tablename = dbo.__tablename__

            query = self._rdb.table(tablename)
            query = query.get(str(dbo.id))
            query = query.delete()

            query.run(conn)

    async def get_count(self, dbo_cls: Type[RethinkDBModelDBOType], attrs: Dict[str, Any] | None = None) -> int:
        with self.connect() as conn:
            tablename = dbo_cls.__tablename__

            query = self._rdb.table(tablename)
            if attrs:
                for key in attrs.keys():
                    attrs[key] = self.adapt_value(attrs[key])
                query = query.filter(attrs)
            query = query.count()

            result = query.run(conn)

            return result

    async def rename_table(self, old_name: str, new_name: str):
        result = None

        with self.connect() as conn:
            query = self._rdb.table(old_name).config().update({
                "name": new_name,
            })

            try:
                result = query.run(conn)
                print(f"Success: Rename table: [{old_name}] => [{new_name}]")
            except:
                print(f"Error: Rename table: [{old_name}] => [{new_name}]")
            return result

    async def check_db(self) -> bool:
        result = []

        with self.connect() as conn:
            table_list = self._rdb.table_list().run(conn)
            for table_name in table_list:
                table_status = self._rdb.table(table_name).status().run(conn)
                conditions = [
                    table_status.get("status").get("all_replicas_ready"),
                    table_status.get("status").get("ready_for_outdated_reads"),
                    table_status.get("status").get("ready_for_reads"),
                    table_status.get("status").get("ready_for_writes"),
                ]

                local_result = all(conditions)
                result.append(local_result)

        return all(result)



from datetime import datetime
from typing import TypeVar

from pydantic import UUID4

from .base import BaseDBO


class RethinkDBBaseDBO(BaseDBO):
    __tablename__: str = "objects"
    __indexes__: list[str] = []

    id: UUID4 | None = None
    created_dt: datetime | None = None
    updated_dt: datetime | None = None

    @classmethod
    def get_fields(cls) -> list[str]:
        return [ field_name for field_name, _ in cls.model_fields.items() ]


ModelDBOType = TypeVar('ModelDBOType', bound=RethinkDBBaseDBO)



from typing import Type, Generic, Any, Dict

from pydantic_yaml import parse_yaml_file_as
from uuid import UUID as base_uuid_type

from ctesiphon.dbo.base import ModelDBOType
from ctesiphon.drivers.data.base import BaseDBODriver
from ctesiphon.dto.base import ModelDTOType


class ModelRepository(Generic[ModelDTOType, ModelDBOType]):
    def __init__(self, driver: BaseDBODriver):
        self.driver = driver

    @property
    def model_dbo(self) -> Type[ModelDBOType]:
        return self.__orig_bases__[0].__args__[1]
    
    @property
    def model_dto(self) -> Type[ModelDTOType]:
        return self.__orig_bases__[0].__args__[0]

    def yaml_to_dto(self, filename: str) -> ModelDTOType:
        return parse_yaml_file_as(self.model_dto, filename)
    
    def yaml_to_dto_list(self, filename: str) -> list[ModelDTOType]:
        return parse_yaml_file_as(list[self.model_dto], filename)

    def dto_to_dbo(self, item: ModelDTOType) -> ModelDBOType:
        item_dict = self.item_to_dict(item)
        return self.model_dbo(**item_dict)

    def dbo_to_dto(self, item: ModelDBOType) -> ModelDTOType:
        item_dict = self.item_to_dict(item)
        return self.model_dto(**item_dict)

    def item_to_dict(self, item: Type[ModelDBOType] | Type[ModelDTOType]) -> dict:
        result = {}

        if item.id is not None:
            result['id'] = item.id

        fields_list_dto = self.model_dto.get_fields()
        fields_list_dbo = self.model_dbo.get_fields()
        fields_list = [ item for item in fields_list_dto if item in fields_list_dbo ]

        for field in fields_list:
            if field not in result:
                result[field] = getattr(item, field)

        return result

    async def init(self):
        await self.driver.init_model(self.model_dbo)

    async def save(self, item: ModelDTOType | dict) -> ModelDTOType:
        if isinstance(item, dict):
            dbo_obj = self.model_dbo(**item)
        else:
            dbo_obj = self.dto_to_dbo(item)
        dbo_obj = await self.driver.save(dbo_obj)

        return self.dbo_to_dto(dbo_obj)

    async def get_by_attr(self, attr: str, value: Any) -> ModelDTOType | None:
        item = await self.driver.get(self.model_dbo, attr, value)

        if not item:
            return None

        return self.dbo_to_dto(item)

    async def get_by_params(
        self,
        attrs: Dict[str, Any] | None = None,
        order_by: str = "id",
        order_direction: str = "desc",
        limit: int = 100,
        offset: int = 0
    ) -> ModelDTOType | None:
        items_result = await self.driver.get_list(
            self.model_dbo,
            attrs,
            order_by=order_by,
            order_direction=order_direction,
            limit=limit,
            offset=offset,
        )

        if len(items_result) == 0:
            return None

        return self.dbo_to_dto(items_result[0])

    async def get_by_id(self, item_id: base_uuid_type | str) -> ModelDTOType | None:
        if isinstance(item_id, base_uuid_type):
            prep_item_id = item_id
        else:
            prep_item_id = base_uuid_type(item_id)

        item = await self.driver.get(self.model_dbo, "id", prep_item_id)

        if not item:
            return None

        return self.dbo_to_dto(item)

    async def get_list_by_attr(
        self,
        attr: str,
        value: Any,
        order_by: str = "id",
        order_direction: str = "desc",
        limit: int = 100,
        offset: int = 0
    ) -> list[ModelDTOType]:
        return await self.get_list_by_params(
            { attr: value },
            order_by=order_by,
            order_direction=order_direction,
            limit=limit,
            offset=offset,
        )

    async def get_list_by_params(
        self,
        attrs: Dict[str, Any] | None = None,
        order_by: str = "id",
        order_direction: str = "desc",
        limit: int = 100,
        offset: int = 0
    ) -> list[ModelDTOType]:
        items_result = await self.driver.get_list(
            self.model_dbo,
            attrs,
            order_by=order_by,
            order_direction=order_direction,
            limit=limit,
            offset=offset,
        )

        return [ self.dbo_to_dto(item) for item in items_result ]
    
    async def get_count_by_params(self, attrs: Dict[str, Any] | None = None) -> int:
        return await self.driver.get_count(self.model_dbo, attrs)

    async def delete(self, item: ModelDTOType) -> None:
        dbo_obj = self.dto_to_dbo(item)
        await self.driver.delete(dbo_obj)
